class AsynchroneousCallback(object):
    """Asynchroneous callbacks are objects which when called in turn call a
    function (or any callable) with a given set of arguments.

    >>> def print_addition(a, b):
    ...     print a + b
    ... 
    >>> print_addition(1, 2)
    3
    >>> cb = AsynchroneousCallback(print_addition, 5)
    >>> cb(5)
    10
    >>> cb(1)
    6
    >>> def rot13_kwargs(**d):
    ...     return dict((k.encode("rot13"), d[k]) for k in d)
    ... 
    >>> rot13_kwargs(foo="bar")
    {'sbb': 'bar'}
    >>> cb = AsynchroneousCallback(rot13_kwargs, sbb="bar")
    >>> cb()
    {'foo': 'bar'}
    >>> cb(quux="lax")
    {'foo': 'bar', 'dhhk': 'lax'}
    """

    __slots__ = "callback", "data"

    def __init__(self, callback, *args, **kwargs):
        """Set callback to "callback", and set the default arguments to any
        other arguments provided.
        """

        self.callback = callback
        self.data = args, kwargs

    def call(self, *args, **kwargs):
        """Call the callback callable with the arguments given at
        construction, with any extra arguments added. For the keyword
        arguments, any keyword arguments passed will override the ones passed
        to the constructor.
        """

        default_args, default_kwargs = self.data
        call_args = default_args + args
        call_kwargs = default_kwargs.copy()
        call_kwargs.update(kwargs)
        return self.callback(*call_args, **call_kwargs)

    __call__ = call  # Convenience

def test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    test()



"""Follow a file line-by-line."""

class Tail(object):
    refresh_rate = 1

    def __init__(self, filename):
        self.fp = file(filename, "r")
        self.fp.seek(0, 2)  # Seek to end.
        self.insts = []
        self.buffer = ""

    def __call__(self, inst):
        self.insts.append(inst)
        return self

    def timer(self):
        d = self.fp.read()
        self.buffer += d
        lines = self.buffer.split("\n")
        self.buffer = lines.pop()
        [self.handle(line) for line in lines]

    def log(self, line):
        [inst.log(line) for inst in self.insts]

    handle = log



#!/usr/bin/env python
"""Block device reader for Linux 2.6"""

from os import path, listdir
from itertools import izip

class BlockStat(object):
    _val_keys = ("read_ios", "read_merges", "read_sectors", "read_ticks",
        "write_ios", "write_merges", "write_sectors", "write_ticks",
        "in_flight", "io_ticks", "time_in_queue")

    def __init__(self, devname):
        self.path = path.join("/", "sys", "block", devname, "stat")

    def _get_vals(self):
        f = file(self.path, "r")
        try:
            d = f.read()
        finally:
            f.close()
        return [int(x) for x in d.replace("\n", "").split(" ") if x]

    def stats(self):
        return dict(izip(self._val_keys, self._get_vals()))

def get_statables():
    base = path.join("/", "sys", "block")
    return iter(listdir(base))
    # So far, this hasn't been needed.
    #for devname in listdir(base):
    #    fn = path.join(base, devname)
    #    if path.isdir(fn) and path.exists(path.join(fn, "stat")):
    #        yield devname

def pprint_all():
    from pprint import pprint

    [pprint((dev, BlockStat(dev).stats())) for dev in get_statables()]

if __name__ == "__main__":
    pprint_all()



"""Memory bar, shows memory usage."""

from pysysmon.readers.meminfo import get_memory_info

class MemoryBar(object):
    name = "Mem"
    refresh_rate = 5

    def __init__(self, name=name):
        self.name = name

    def max(self):
        if not hasattr(self, "_maxmem"):
            self._maxmem = get_memory_info()["MemTotal"]
        return self._maxmem

    def current(self):
        d = get_memory_info()
        return self.max() - d["MemFree"] - d["Buffers"] - d["Cached"]

    def timer(self):
        return True  # Refresh this bar.



"""Main module, run me."""

import sys
from os import path, getcwd
import imp
import signal
import curses
import getopt

from gobject import MainLoop, io_add_watch, timeout_add, IO_IN
from pysysmon import ui

def get_conf(fn):
    fp = file(fn, "U")
    try:
        m = imp.load_module("<pysysmon configuration>", fp, fn,
            ("py", "U", imp.PY_SOURCE))
    finally:
        fp.close()
    return m

def usage():
    print >>sys.stderr, "usage: %s [-c <conf>]" % sys.argv[0]
    sys.exit(1)

def no_conf(m=None):
    pdir = path.dirname(__file__)

    print >>sys.stderr, "No usable configuration found."
    print >>sys.stderr
    print >>sys.stderr, "Perhaps you need to set one up?"
    print >>sys.stderr, "See %s/pysysmon_conf.py.example." % (pdir,)
    if m:
        print >>sys.stderr
        print >>sys.stderr, str(m)
    sys.exit(2)

def main():
    conf_path = "~/.pysysmon_conf.py"
    try:
        opts, args = getopt.getopt(sys.argv[1:], "c:")
    except getopt.GetoptError:
        usage()
    if args:
        usage()

    for opt, val in opts:
        if opt == "-c":
            conf_path = val

    try:
        conf = get_conf(path.expanduser(conf_path))
    except IOError, e:
        no_conf(e)

    def inner(s):
        z = ui.UI(s, conf)
        for cb, timer in z.get_timers():
            timeout_add(timer * 1000, cb)
        for fp, cb in z.get_read_watches():
            io_add_watch(fp, IO_IN, cb)
        while True:
            ml = MainLoop()
            signal.signal(signal.SIGINT, lambda s, f: ml.quit())
            try:
                ml.run()
            except KeyboardInterrupt:
                z.log("KeyboardInterrupt from mainloop")
                continue
            else:
                break
        print
    curses.wrapper(inner)

if __name__ == "__main__":
    main()



"""Test bar."""

class TestBar(object):
    name = "test"
    refresh_rate = 1
    max = lambda s: 100
    timer = lambda s: True

    def current(self):
        self.i = (getattr(self, "i", 0) + 1) % 101
        return self.i



"""Lighttpd access log parser"""

import re
from time import sleep, strptime, strftime
from pysysmon.utils import readable_size
from pysysmon.loggers.tail import Tail
from pysysmon.ui import attr_escape

line_re = re.compile(r'^([^ ]+) ([^ ]+) ([^ ]+) \[([^\]]+)\] "([A-Z]+) ([^ ]+) ([^"]+)" (\d+) (\d+|\-) "(\-|[^"]+)" "([^"]+)"')
time_re = re.compile(r"^(..)/(...)/(....):(........)")

class LighttpdLog(Tail):
    default_colors = {"green": 2, "yellow": 3, "red": 1,}
    max_omissions = 5

    user_agents = ("Firefox", "BonEcho", "Opera", "MSIE 5", "MSIE 6",
        "MSIE 7", "MSIE", "Safari", "wget", "Camino")
    os_names = ("Linux", "Windows", "FreeBSD", "Macintosh", "Googlebot")

    def __init__(self, filename, colors={}):
        super(LighttpdLog, self).__init__(filename)
        self.colors = self.default_colors.copy()
        self.colors.update(colors)
        self.last_ip = None
        self.prints_omitted = 0

    def handle(self, line):
        mo = line_re.search(line)
        if mo:
            line = self.get_formatted(*mo.groups())
        self.log(line)

    def get_formatted(self, ip, http_host, user, timestamp, verb, path,
            httpver, statuscode, length, referrer, useragent):
        # Get status color of line.
        color = "default"
        if statuscode.startswith("2"):
            color = "green"  # 200 OK, e.g.
        elif statuscode.startswith("3"):
            color = "yellow"
        elif statuscode.startswith("4") or statuscode.startswith("5"):
            color = "red"
        # Get printed IP address.
        print_ip = ip
        if ip == self.last_ip:
            print_ip = " " * len(print_ip)
            self.prints_omitted += 1
            if self.prints_omitted > self.max_omissions:
                self.prints_omitted = 0
                print_ip = ip
        self.last_ip = ip
        # Format HTTP host
        if http_host == "-":
            http_host = ""
        else:
            http_host = " %s" % (http_host,)
        # Length
        print_length = length
        if length != "-":
            if length == "0":
                print_length = "0 B"
            else:
                print_length = readable_size(length)
        print_length = " %s " % print_length
        # Get timestamp.
        ts_mo = time_re.match(timestamp)
        print_ts = timestamp
        if ts_mo:
            day, month, year, time = ts_mo.groups()
            day = "%2d" % int(day)
            print_ts = " ".join((month, day, time))
        # Get additional info.
        addit = []
        if user != "-":
            addit.append("user: %s" % attr_escape(user))
        if referrer != "-":
            addit.append("ref: %s" % attr_escape(referrer))
        # Like rudimentary heuristics for UA name and OS name.
        if useragent != "-":
            matched_uas = []
            for ua in self.user_agents:
                if ua in useragent:
                    matched_uas.append(ua)
            matched_os = []
            for os in self.os_names:
                if os in useragent:
                    matched_os.append(os)
            if matched_uas:
                addit.append("agent: " + attr_escape(", ".join(matched_uas)))
            else:
                addit.append("agent?: " + attr_escape(useragent))
            if matched_os:
                addit.append("os: " + attr_escape(", ".join(matched_os)))
        # Nicely format addit.
        print_addit = ""
        if addit:
            print_addit = "(%s)" % (", ".join(addit),)
        # Then return the whole line
        return "%s lighty %s%s %s %%%d%s%%0%s%s" % (print_ts,
            print_ip, attr_escape(http_host), attr_escape(verb),
            self.colors.get(color, 9), attr_escape(path), print_length,
            print_addit)



"""Network interface service"""

from os import path
import fcntl
from struct import pack, unpack
import socket

SIOCGIFFLAGS = 0x8913
IFF_UP = 0x1

#struct ifreq32 {
#        union {
#                char    ifrn_name[16];
#        } ifr_ifrn;
#        union {
#                struct  sockaddr ifru_addr;
#                struct  sockaddr ifru_dstaddr;
#                struct  sockaddr ifru_broadaddr;
#                struct  sockaddr ifru_netmask;
#                struct  sockaddr ifru_hwaddr;
#                short   ifru_flags;
#                int     ifru_ivalue;
#                int     ifru_mtu;
#                struct  ifmap32 ifru_map;
#                char    ifru_slave[16];   /* Just fits the size */
#                char    ifru_newname[16];
#                __kernel_caddr_t32 ifru_data;
#        } ifr_ifru;
#};

def _get_flags_for(iface):
    format = "16s H 16x"  # ifrn_name, ifru_flags + padding
    # Useless socket, because Linux wants it.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    req = pack(format, iface, 0)
    resp = fcntl.ioctl(sock, SIOCGIFFLAGS, req)
    return unpack(format, resp)[1]

def is_up(iface):
    try:
        return _get_flags_for(iface) & IFF_UP > 0
    except IOError:
        return False

class InterfaceUp(object):
    refresh_rate = 60

    def __init__(self, iface, name=None):
        self.name = name or iface
        self.iface = iface

    def ok(self):
        return is_up(self.iface)



#!/usr/bin/env python
from distutils.core import setup

setup(name="pysysmon", version="0.0.25",
    url="http://lericson.se/",
    author="Ludvig Ericson", author_email="ludvig@lericson.se",
    description="A curses-based system monitor.",
    packages=["pysysmon", "pysysmon.readers", "pysysmon.services",
        "pysysmon.bars", "pysysmon.loggers"], scripts=["pysysmon/pysysmon"],
    package_data={"pysysmon": ["pysysmon_conf.py.example"]})



"""Memory info reader for Linux."""

from os import path

def get_memory_info():
    f = file(path.join("/", "proc", "meminfo"))
    d = {}
    try:
        for l in f:
            k, rv = l.split(": ", 1)
            rv = rv.lstrip(" ")[:-1]  # Trim off \n too.
            if not rv.endswith(" kB"):
                raise ValueError("line ends not in ' kb'.")
            d[k] = int(rv[:-3]) * 1024  # it's KiB => bytes
    finally:
        f.close()
    return d

def pprint_all():
    from pprint import pprint
    pprint(get_memory_info())

if __name__ == "__main__":
    pprint_all()
